#include "xparameters.h"
#include "xio.h"
#include "xintc.h"
#include "xintc_l.h"

#include "ftBlock.h"

#define POP_SIZE			64		// The Population size
#define TOURNAMENT_SIZE	4		// The number of individuals used in tournament selection
#define MUTATION_RATE	2		// Mutate MUTATION_RATE out of 256 times.
#define MAX_GEN			500	// Maximum number of generations per run

#define MAX_FITNESS		0		// TODO Max fitness

// Prototypes
void loadTargetTruthTable();
void outputTargetTruthTable();
void randomisePopulation(ArrayConfig *pop, Xuint32 popSize);
void mutateArrayConfig(ArrayConfig *source, ArrayConfig *dest, Xuint32 mutationRate);
Xuint32 tournamentSelect(ArrayConfig* population, Xuint32 popSize, Xuint32 tournamentSize);
Xuint32 calcFitness(Xuint32 matchData);

void buttonPressISR(void);
void setupInterrupts(void);

Xuint32 findSolution(Xuint32 arrayNum);

// Globals
ArrayConfig pop1[POP_SIZE];
ArrayConfig pop2[POP_SIZE];

ArrayConfig *population;
ArrayConfig *nextPopulation;
ArrayConfig *tempPopulation;

int main(void)
{
	print("\nStarting Experiment 1\n");

	// Load the EvoBlock Target RAM with the target truth table
	// This should be for a 4 input parity generator, with both even and odd parity outputs
	loadTargetTruthTable();

	// Display RAM contents
	outputTargetTruthTable();

	// STOP 1
	return 0;

	// Switch Cell Array vector input to program control (external vector source)
	FTBLOCK_writeInputSel(0x7);

	// Evolve a solution for Cell Array 0
	Xuint32 result = findSolution(0);
	if (result)
	{
		xil_printf("Array 0: No solution after %d generations\n", MAX_GEN);
		return 1;
	}

	// Evolve a solution for Cell Array 1
	result = findSolution(1);
	if (result)
	{
		xil_printf("Array 1: No solution after %d generations\n", MAX_GEN);
		return 1;
	}

	// Evolve a solution for Cell Array 2
	result = findSolution(2);
	if (result)
	{
		xil_printf("Array 2: No solution after %d generations\n", MAX_GEN);
		return 1;
	}

	// STOP 2
	return 0;

	// Switch to Internal Vector Source
	FTBLOCK_writeInputSel(0x0);

	// Reset Error Detection Flags
	FTBLOCK_resetError(0);

	// Setup and enable Interrupts
	setupInterrupts();
	microblaze_enable_interrupts();

	print("\nReady for fault injection\n");

	Xuint32 error;
	Xuint32 arrayNum;

	while(1)
	{
		// Check for an error
		error = FTBLOCK_readError() & 0x7;

		// If an error is found, evolve a new solution to recover
		if (error!=0)
		{
			// Find out which array is faulty
			if (error&0x1) arrayNum = 0;
			else if (error&0x2) arrayNum = 1;
			else if (error&0x4) arrayNum = 2;

			// Switch Cell Array vector input to program control (external vector source)
			FTBLOCK_writeInputSel(0x7);

			// Find a new working solution
			result = findSolution(arrayNum);
			if (result)
			{
				xil_printf("Array 0: No solution after %d generations\n", MAX_GEN);
				return 1;
			}

			// Switch back to the internal vector source
			FTBLOCK_writeInputSel(0x0);

			// Reset the error detection flags
			FTBLOCK_resetError(0);
		}
	}
}

Xuint32 findSolution(Xuint32 arrayNum)
{
	// Mak sure we have a valid array number
	if (arrayNum > 2) arrayNum = 2;

	xil_printf("Evolving Solution in Array %d\n", arrayNum);

	Xuint32 p = 0;
	Xuint32 gen = 0;
	Xuint32 vector = 0;
	Xuint32 outputMatchData;
	Xuint32 match;
	Xuint32 fitness;
	Xuint32 fittest;
	Xuint32 parent;
	Xuint32 tournament;
	Xuint32 searching;

	// Initialise Populations
	population = pop1;
	nextPopulation = pop2;
	randomisePopulation(population, POP_SIZE);

	gen = 0;
	searching = 1;

	while(searching)
	{
		// Iterate through the population
		for (p=0 ; p<POP_SIZE ; p++)
		{
			// Reset individual's fitness to 0
			population[p].fitness = 0;

			// Configure the Cell Array
			configureArray(arrayNum, &population[p]);

			// Iterate through input vectors, 16 vectors as there are 4 inputs (2^4=16)
			for (vector=0; vector<16 ; vector++)
			{
				// Apply the vector to the array
				FTBLOCK_writeVector(vector);

				// Read the output and match data
				outputMatchData = FTBLOCK_readOutputMatch();

				// Extract the match data for this array
				match = outputMatchData>>6;		// Shift away output data, leaving match data
				match = match >> (arrayNum*2);	// Shift array match data to LSB
				match = match & 0x3;					// Mask out unwatned bits

				// Calculate and add fitness for this vector
				population[p].fitness += calcFitness(match);
			}
		}

		// Find the fittest
		fittest = 0;
		for (p=0 ; p<POP_SIZE ; p++) if (population[p].fitness >= population[fittest].fitness) fittest = p;

		// Check if solution found
		if (population[fittest].fitness == MAX_FITNESS)
		{
			xil_printf("Solution Found, Gen: %d, fittest: %3d (%3d)\n", gen, fittest, population[fittest].fitness);
			searching = 0;
			continue;
		}

		// Show Progress
		xil_printf("Gen: %3d, fittest: %3d (%3d), Config: %03X\n",
						gen, fittest, population[fittest].fitness, population[fittest].cellConfigs[0]);

		// If reached the generation limit, end this run
		if (gen == MAX_GEN)
		{
			xil_printf("Solution Not found, Gen: %d, fittest: %3d (%3d)\n", gen, fittest, population[fittest].fitness);
			return 1;
		}

		// Elitism
		nextPopulation[0] = population[fittest];

		// Create next population
		for (tournament=1 ; tournament<POP_SIZE ; tournament++)
		{
			// Use Tournament Selection to get one individual
			parent = tournamentSelect(population, POP_SIZE, TOURNAMENT_SIZE);
			mutateArrayConfig(&population[parent], &nextPopulation[tournament], MUTATION_RATE);
		}

		// Swap Populations
		tempPopulation = population;
		population = nextPopulation;
		nextPopulation = tempPopulation;

		// Increment the generation number
		gen++;
	}

	// Configure the Cell Array with the fittest solution
	configureArray(arrayNum, &population[fittest]);

	return 0;
}

// ----------------------------------------------------------------------------------------------------------
// TODO loadTargetTruthTable
// Load the target RAM with the target truth table
// ----------------------------------------------------------------------------------------------------------
void loadTargetTruthTable()
{
	// Initialise vector to 0
	// This will automatically increments with every write to the Target RAM
	FTBLOCK_writeVector(0);

	// Load the Target RAM using FTBLOCK_writeTargetRAM(data);
}

// ----------------------------------------------------------------------------------------------------------
// outputTargetTruthTable
// Display the RAM contents
// ----------------------------------------------------------------------------------------------------------
void outputTargetTruthTable()
{
	Xuint32 vector;
	Xuint32 output;
	Xuint32 even;
	Xuint32 odd;

	char buffer[5];

	// Iterate though all test vectors loading the correct output value
	for(vector=0 ; vector<16 ; vector++)
	{
		// Set the RAM address (vetor)
		FTBLOCK_writeVector(vector);

		// Calculate parity data
		output = FTBLOCK_readTargetRAM() & 0x3;
		even = output & 0x1;
		odd = (output >> 1) & 0x1;

		Xuint32 b, p=8;
		for (b=0; b<4 ; b++)
		{
			buffer[b] = ((vector & p) == 0) ? '0' : '1';
			p >>= 1;
		}

		// Null Terminate buffer
		buffer[4] = '\0';

		xil_printf("RAM[%2d]=%s, odd:%d, even:%d\n", vector, buffer, odd, even);
	}
}

// ----------------------------------------------------------------------------------------------------------
// randomisePopulation
// Iterate through the population randomizin each individual
// ----------------------------------------------------------------------------------------------------------
void randomisePopulation(ArrayConfig *pop, Xuint32 popSize)
{
	Xuint32 individual;

	// Iterate though the population
	for(individual=0 ; individual<popSize ; individual++)
	{
		// Randomise the array configuration
		randomiseArrayConfig(&pop[individual]);
	}
}

// ----------------------------------------------------------------------------------------------------------
// tournamentSelect
// Select 'tournamentSize' individuals randomly from 'population', return the number of the fittest
//	Note: Contenders can the same
// ----------------------------------------------------------------------------------------------------------
Xuint32 tournamentSelect(ArrayConfig* population, Xuint32 popSize, Xuint32 tournamentSize)
{
	Xuint32 p;
	Xuint32 randNum;
	Xuint32 contender[tournamentSize];

	// Select random contenders
	for (p=0 ; p<tournamentSize ; p++)
	{
		// Get a random individual number
		randNum = rand() % popSize;

		// Store contender number
		contender[p] = randNum;
	}

	// Find Fittest Contender
	Xuint32 fittest = contender[0];

	for (p=0 ; p<tournamentSize ; p++)
	{
		if (population[contender[p]].fitness > population[fittest].fitness) fittest = contender[p];
	}

	return fittest;
}

// ----------------------------------------------------------------------------------------------------------
// TODO mutateArrayConfig
// Iterate through cell configs, flip config bit if randum number (0 to 255) is less that MUTATION_RATE
// ----------------------------------------------------------------------------------------------------------
void mutateArrayConfig(ArrayConfig *source, ArrayConfig *dest, Xuint32 mutationRate)
{
}

// ----------------------------------------------------------------------------------------------------------
// TODO calcFitness
// Sum the number of '1' in the matchData (This is the number of matching bits)
// ----------------------------------------------------------------------------------------------------------
Xuint32 calcFitness(Xuint32 matchData)
{
	return 0;
}

// ----------------------------------------------------------------------------------------------------------
// buttonPressISR
// Interrupt Service Routine for button presses
// ----------------------------------------------------------------------------------------------------------
void buttonPressISR(void)
{
	// Get which button was pressed
	Xuint32 buttons = XIo_In32(XPAR_OPB_BUTTONS_0_BASEADDR + 4);
	XIo_Out32(XPAR_OPB_BUTTONS_0_BASEADDR + 4, 0);

	// Decode Array Number
	Xuint32 arrayNum;

	if (buttons&0x1) arrayNum = 0;
	else if (buttons&0x2) arrayNum = 1;
	else if (buttons&0x4) arrayNum = 2;
	else return;

	// Selection Random cell (0->31)
	Xuint32 cellNum = rand() & 0x1F;

	// Select random fail mode (1->3)
	Xuint32 mode = (rand()%3) + 1;

	// Inject Fault
	FTBLOCK_writeFailModeArrayNum(arrayNum);
	FTBLOCK_writeFailModeCellNum(cellNum);
	FTBLOCK_writeFailMode(mode);

	xil_printf("Failing Array: %d, cell: %d ", arrayNum, cellNum);

	switch (mode)
	{
		case 1:	print("Invert\n"); break;
		case 2:	print("SA0\n"); break;
		case 3:	print("SA1\n"); break;
		default:	print("Mode Error\n"); break;
	}
}

void setupInterrupts(void)
{
	// Register Button Interrupt Handler
	XIntc_RegisterHandler(  XPAR_OPB_INTC_0_BASEADDR,
									XPAR_OPB_INTC_0_OPB_BUTTONS_0_PRESSIRQ_INTR,
									(XInterruptHandler)buttonPressISR, (void *)0);

	// Enable the Interrupt Controller
	XIntc_mMasterEnable(XPAR_OPB_INTC_0_BASEADDR);

	// Set Interrupt Controller Enable for the buttons
	XIntc_mEnableIntr(XPAR_OPB_INTC_0_BASEADDR, XPAR_OPB_BUTTONS_0_PRESSIRQ_MASK);

	// Enable Microblaze Interrupts
	microblaze_enable_interrupts();
}
