library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity FTBlock is
	port (	clk					: in	std_logic;
				reset					: in	std_logic;

				configArrayNum		: in	integer range 0 to 2;
				configCellNum		: in	integer range 0 to 31;
				config				: in	std_logic_vector(11 downto 0);
				configWEn			: in	std_logic;

				outputSel0			: in	integer range 0 to 31;
				outputSel1			: in	integer range 0 to 31;
				outputSelWEn		: in	std_logic;

				failModeArrayNum	: in	integer range 0 to 2;
				failModeCellNum	: in	integer range 0 to 31;
				failMode				: in	std_logic_vector(1 downto 0);
				failModeWEn			: in	std_logic;

				ramDIn				: in  std_logic_vector(1 downto 0);
				ramWEn				: in  std_logic;
				ramDOut				: out std_logic_vector(1 downto 0);

				inputSel				: in	std_logic_vector(2 downto 0);

				vector				: in	std_logic_vector(3 downto 0);
				output				: out	std_logic_vector(5 downto 0);
				match					: out	std_logic_vector(5 downto 0);
				arrayError			: out	std_logic_vector(2 downto 0));

end FTBlock;

architecture Arch of FTBlock is

	type RAM_TYPE is array (15 downto 0) of std_logic_vector(1 downto 0);
	type OUTPUT_TYPE is array (2 downto 0) of std_logic_vector(1 downto 0);

	component CellArray is
		port (	clk					: in	std_logic;
					reset					: in	std_logic;
					dIn					: in	std_logic_vector(3 downto 0);
					dOut					: out	std_logic_vector(1 downto 0);
					configCellNum		: in	integer range 0 to 31;
					config				: in	std_logic_vector(11 downto 0);
					configWEn			: in	std_logic;
					outputSel0			: in	integer range 0 to 31;
					outputSel1			: in	integer range 0 to 31;
					outputSelWEn		: in	std_logic;
					failModeCellNum	: in	integer range 0 to 31;
					failMode				: in	std_logic_vector(1 downto 0);
					failModeWEn			: in	std_logic);
	end component;

	signal targetRAM			: RAM_TYPE;
	signal targetRAMOut		: std_logic_vector(1 downto 0);

	signal internalVector	: std_logic_vector(3 downto 0);
	signal arrayOutReg		: OUTPUT_TYPE;

	signal votedOut			: std_logic_vector(1 downto 0);

	type INTERNAL_TEST is (APPLY, DELAY1, DELAY2, VOTE, CHECK);
	signal state				: INTERNAL_TEST;

begin

ramDOut <= targetRAMOut;

-- TargetRAM --------------------------------------------------------------------------------------------------
-- Target truth table
---------------------------------------------------------------------------------------------------------------
process (clk)
begin
	if (clk'event and clk='1') then
		if (ramWEn = '1') then
			targetRAM(to_integer(unsigned(vector))) <= ramDIn;
			targetRAMOut <= ramDIn;
		else
			targetRAMOut <= targetRAM(to_integer(unsigned(vector)));
		end if;
	end if;
end process;

-- InternalVectorSource ---------------------------------------------------------------------------------------
-- Creates an internal vector source, that is once solutions have been evolved
---------------------------------------------------------------------------------------------------------------
InternalVectorSource : process(clk)
begin
	if (clk'event and clk='1') then
		if (reset='1') then
			internalVector <= (others => '0');
			state <= APPLY;

			votedOut <= (others => '0');
--			tmrError	<= '0';
			arrayError <= (others => '0');

		else
			case state is
				when APPLY =>
					internalVector <= std_logic_vector(unsigned(internalVector) + 1);
					state <= DELAY1;

				when DELAY1 =>
					state <= DELAY2;

				when DELAY2 =>
					state <= VOTE;

				when VOTE =>
					votedOut <= (arrayOutReg(0) and arrayOutReg(1)) or (arrayOutReg(1) and arrayOutReg(2)) or (arrayOutReg(0) and arrayOutReg(2));
					state <= CHECK;

				when CHECK =>
--					if (votedOut/=targetRAMOut) then
--						tmrError <= '1';
--					else
--						tmrError <= '0';
--					end if;
					
					arrayError <= (others => '0');

					if (arrayOutReg(0)/=votedOut and inputSel(0)='0') then
						arrayError(0) <='1';
					end if;

					if (arrayOutReg(1)/=votedOut and inputSel(1)='0') then
						arrayError(1) <='1';
					end if;

					if (arrayOutReg(2)/=votedOut and inputSel(2)='0') then
						arrayError(2) <='1';
					end if;

					state <= APPLY;
			end case;

		end if;
	end if;
end process;

-- ArrayGenerate ----------------------------------------------------------------------------------------------
-- Create a block of 3 Cell Arrays
---------------------------------------------------------------------------------------------------------------
ArrayGenerate : for arrayNum in 0 to 2 generate

	signal arrayIn				: std_logic_vector(3 downto 0);
	signal arrayOut			: std_logic_vector(1 downto 0);

	signal arrayOutputSelWEn	: std_logic;
	signal arrayFailModeWEn		: std_logic;
	signal arrayConfigWEn		: std_logic;

begin

	-- Input Mulitplexer
	arrayIn <= internalVector when (inputSel(arrayNum)='0') else vector;

	arrayConfigWEn		<= '1' when (configArrayNum=arrayNum and configWEn='1') else '0';
	arrayOutputSelWEn	<= '1' when (configArrayNum=arrayNum and outputSelWEn='1') else '0';
	arrayFailModeWEn	<= '1' when (failModeArrayNum=arrayNum and failModeWEn='1') else '0';

	CellArrayInst : CellArray
	port map(clk					=> clk,
				reset					=> reset,
				dIn					=> arrayIn,
				dOut					=> arrayOut,
				configCellNum		=> configCellNum,
				config				=> config,
				configWEn			=> arrayConfigWEn,
				outputSel0			=> outputSel0,
				outputSel1			=> outputSel1,
				outputSelWEn		=> arrayOutputSelWEn,
				failModeCellNum	=> failModeCellNum,
				failMode				=> failMode,
				failModeWEn			=> arrayFailModeWEn);

	process (clk)
	begin
		if (clk'event and clk='1') then
			arrayOutReg(arrayNum) <= arrayOut;
		end if;
	end process;

	output((arrayNum*2)+1 downto (arrayNum*2)) <= arrayOutReg(arrayNum);
	match((arrayNum*2)+1 downto (arrayNum*2)) <= arrayOutReg(arrayNum) xnor targetRAMOut; 

end generate;

end Arch;
